@@ -4,7 +4,6 @@ from typing import List, Union
 import httpx
 import orjson
 import re
-import pendulum
 from authlib.oauth2.rfc6749 import OAuth2Token
 from networkx import DiGraph
 from authlib.integrations.httpx_client import OAuth2Client
@@ -21,8 +20,8 @@ import fhir.resources
 
 from fhir_kindling.fhir_query import FHIRQuerySync, FHIRQueryAsync
 from fhir_kindling.fhir_query.query_response import QueryResponse
-from fhir_kindling.fhir_query.query_parameters import FHIRQueryParameters, FieldParameter, QueryOperators
-from fhir_kindling.fhir_server.auth import generate_auth, BearerAuth, auth_info_from_env
+from fhir_kindling.fhir_query.query_parameters import FHIRQueryParameters
+from fhir_kindling.fhir_server.auth import BearerAuth, auth_info_from_env
 from fhir_kindling.fhir_server.server_responses import ResourceCreateResponse, BundleCreateResponse, ServerSummary, \
     TransferResponse
 from fhir_kindling.util.references import check_missing_references, reference_graph
@@ -30,9 +29,18 @@ from fhir_kindling.util.references import check_missing_references, reference_gr
 
 class FhirServer:
 
-    def __init__(self, api_address: str, username: str = None, password: str = None, token: str = None,
-                 client_id: str = None, client_secret: str = None, oidc_provider_url: str = None,
-                 auth: httpx.Auth = None, headers: dict = None,
+    def __init__(self,
+                 api_address: str,
+                 username: str = None,
+                 password: str = None,
+                 token: str = None,
+                 client_id: str = None,
+                 client_secret: str = None,
+                 oidc_provider_url: str = None,
+                 auth: httpx.Auth = None,
+                 headers: dict = None,
+                 proxies: Union[dict, str] = None,
+                 timeout: int = 5,
                  fhir_server_type: str = "hapi"):
         """
         Initialize a FHIR server connection
@@ -69,6 +77,8 @@ class FhirServer:
 
         self._auth = auth
         self._headers = headers
+        self._proxies = proxies
+        self._timeout = timeout
 
     @classmethod
     def from_env(cls, no_auth: bool = False) -> 'FhirServer':
@@ -227,7 +237,8 @@ class FhirServer:
         """
         if isinstance(reference, Reference):
             reference = reference.reference
-        r = self._sync_client().get(f"{self.api_address}/{reference}")
+        with self._sync_client() as client:
+            r = client.get(f"{self.api_address}/{reference}")
         r.raise_for_status()
         resource_dict = r.json()
         resource = construct_fhir_element(resource_dict["resourceType"], resource_dict)
@@ -267,12 +278,15 @@ class FhirServer:
         """
         str_references = [reference if isinstance(reference, str) else reference.reference for reference in references]
 
-        # resources = self._get_many_query(str_references)
-        # todo use batched/transaction requests to get all resources in one request
         get_many_transaction = self._make_get_many_transaction(str_references)
-        response = self._sync_client().post(self.api_address, json=self._json_dict(get_many_transaction))
-        #
-        entries = response.json()["entry"]
+        with self._sync_client() as client:
+            r = client.post(self.api_address, json=self._json_dict(get_many_transaction))
+            try:
+                r.raise_for_status()
+            except Exception as e:
+                print(r.text)
+                raise e
+        entries = r.json()["entry"]
         resources = [construct_fhir_element(entry["resource"]["resourceType"], entry["resource"]) for entry in entries]
 
         return resources
@@ -428,16 +442,26 @@ class FhirServer:
         """
 
         update_bundle = self._make_update_transaction(resources)
-        response = self._sync_client().post(self.api_address, json=self._json_dict(update_bundle))
-        response.raise_for_status()
-
-        return response.json()
+        with self._sync_client() as client:
+            r = client.post(self.api_address, json=self._json_dict(update_bundle))
+            try:
+                r.raise_for_status()
+            except Exception as e:
+                print(r.text)
+                raise e
+        return r.json()
 
     async def update_async(self, resources: List[Union[FHIRResourceModel, dict]]):
         update_bundle = self._make_update_transaction(resources)
-        response = await self._async_client().post(self.api_address, json=self._json_dict(update_bundle))
-        response.raise_for_status()
-        return response.json()
+
+        async with self._async_client() as client:
+            r = await client.post(self.api_address, json=self._json_dict(update_bundle))
+            try:
+                r.raise_for_status()
+            except Exception as e:
+                print(r.text)
+                raise e
+        return r.json()
 
     def delete(self,
                resources: List[Union[FHIRResourceModel, dict]] = None,
@@ -456,10 +480,14 @@ class FhirServer:
 
         delete_bundle = self._make_delete_transaction(resources, references, query)
 
-        response = self._sync_client().post(self.api_address, json=self._json_dict(delete_bundle))
-        response.raise_for_status()
-
-        return response
+        with self._sync_client() as client:
+            r = client.post(self.api_address, json=self._json_dict(delete_bundle))
+            try:
+                r.raise_for_status()
+            except Exception as e:
+                print(r.text)
+                raise e
+        return r
 
     async def delete_async(self,
                            resources: List[Union[FHIRResourceModel, dict]] = None,
@@ -467,10 +495,15 @@ class FhirServer:
                            query: FHIRQuerySync = None):
         delete_bundle = self._make_delete_transaction(resources, references, query)
 
-        response = await self._async_client().post(self.api_address, json=self._json_dict(delete_bundle))
-        response.raise_for_status()
+        async with self._async_client() as client:
+            r = await client.post(self.api_address, json=self._json_dict(delete_bundle))
+            try:
+                r.raise_for_status()
+            except Exception as e:
+                print(r.text)
+                raise e
 
-        return response
+        return r
 
     def transfer(self, target_server: 'FhirServer', query_result: QueryResponse) -> TransferResponse:
         """
@@ -643,18 +676,26 @@ class FhirServer:
         return entry
 
     def _upload_bundle(self, bundle: Bundle) -> BundleCreateResponse:
-        r = self._sync_client().post(url=self.api_address, json=self._json_dict(bundle))
-        try:
-            r.raise_for_status()
-        except Exception as e:
-            print(r.text)
-            raise e
+        with self._sync_client() as client:
+
+            r = client.post(url=self.api_address, json=self._json_dict(bundle))
+            try:
+                r.raise_for_status()
+            except Exception as e:
+                print(r.text)
+                raise e
         bundle_response = BundleCreateResponse(r, bundle)
         return bundle_response
 
     def _upload_resource(self, resource: Resource) -> httpx.Response:
         url = self.api_address + "/" + resource.get_resource_type()
-        r = self._sync_client().post(url=url, headers=self.headers, auth=self._auth, json=self._json_dict(resource))
+        with self._sync_client() as client:
+            r = client.post(url=url, json=self._json_dict(resource))
+            try:
+                r.raise_for_status()
+            except Exception as e:
+                print(r.text)
+                raise e
         return r
 
     async def _upload_resource_async(self, resource: Resource) -> httpx.Response:
@@ -681,8 +722,13 @@ class FhirServer:
 
     def _get_meta_data(self):
         url = self.api_address + "/metadata"
-        r = self._sync_client().get(url)
-        r.raise_for_status()
+        with self._sync_client() as client:
+            r = client.get(url)
+            try:
+                r.raise_for_status()
+            except Exception as e:
+                print(r.text)
+                raise e
         response = r.json()
         self._meta_data = response
 
@@ -690,6 +736,8 @@ class FhirServer:
         client = httpx.Client(
             headers=self.headers,
             auth=self.auth,
+            proxies=self._proxies,
+            timeout=self._timeout,
         )
         return client
 
@@ -697,31 +745,29 @@ class FhirServer:
         client = httpx.AsyncClient(
             headers=self.headers,
             auth=self.auth,
+            proxies=self._proxies,
+            timeout=self._timeout,
         )
         return client
 
-    # def _setup(self):
-    #     self.auth = self._validate_auth()
-    #     self.session.auth = self.auth
-    #     self.session.headers.update(self.headers)
-
     def _make_server_summary(self) -> ServerSummary:
         resources = []
         summary = {
             "name": self.api_address
         }
-        for resource in self.rest_resources:
-            url = self.api_address + "/" + resource + "?_summary=count"
-            r = self._sync_client().get(url)
-            r.raise_for_status()
+        with self._sync_client() as client:
+            for resource in self.rest_resources:
+                url = self.api_address + "/" + resource + "?_summary=count"
+                r = client.get(url)
+                r.raise_for_status()
 
-            resource_dict = {
-                "resource": resource,
-                "count": r.json().get("total")
-            }
-            resources.append(resource_dict)
+                resource_dict = {
+                    "resource": resource,
+                    "count": r.json().get("total")
+                }
+                resources.append(resource_dict)
 
-        summary["resources"] = resources
+            summary["resources"] = resources
         summary = ServerSummary(**summary)
         return summary
 
@@ -769,43 +815,6 @@ class FhirServer:
         else:
             raise ValueError(f"Malformed API URL: {api_address}")
 
-    def _validate_auth(self) -> Union[httpx.Auth, None]:
-        if self._auth and (self.username or self.password):
-            raise ValueError(
-                "Only one authentication method can be used, auth object and username/password are mutually exclusive")
-        if self._auth and self.token:
-            raise ValueError("Only one authentication method can be used, auth object and token are mutually exclusive")
-        if self._auth and (self.client_id or self.client_secret or self.oidc_provider_url):
-            raise ValueError(
-                "Only one authentication method can be used, auth object and OIDC auth are mutually exclusive")
-
-        if self.token and (self.client_id or self.client_secret or self.oidc_provider_url):
-            raise ValueError("Only one authentication method can be used, token and OIDC auth are mutually exclusive")
-
-        if self.token and (self.username or self.password):
-            raise ValueError(
-                "Only one authentication method can be used, token and username/password are mutually exclusive")
-
-        if (self.username or self.password) and (self.client_id or self.client_secret or self.oidc_provider_url):
-            raise ValueError("Only one authentication method can be used, username/password and "
-                             "OIDC auth are mutually exclusive")
-
-        if self._auth:
-            return self._auth
-        # OIDC authentication
-        if self.client_id:
-            self._get_oidc_token()
-            return generate_auth(token=self.token)
-        # basic or static token authentication
-        elif self.username and self.password:
-            return generate_auth(self.username, self.password)
-
-        # static token auth
-        elif self.token:
-            return generate_auth(token=self.token)
-        else:
-            return None
-
     def _make_get_many_transaction(self, str_references: List[str]):
         get_bundle = Bundle.construct()
         get_bundle.type = "batch"
@@ -816,25 +825,6 @@ class FhirServer:
         get_bundle = Bundle(**get_bundle.dict(exclude_none=True))
         return get_bundle
 
-    def _get_many_query(self, str_references: List[str]) -> List[FHIRAbstractModel]:
-        # todo remove this in favor of batch transactions
-        resource_refs = {}
-        for reference in str_references:
-            resource_type, id = reference.split("/")
-            resource_id_list = resource_refs.get(resource_type, [])
-            resource_id_list.append(id)
-            resource_refs[resource_type] = resource_id_list
-        resources = []
-        for resource_type, resource_ids in resource_refs.items():
-            params = FHIRQueryParameters(
-                resource=resource_type,
-                resource_parameters=[FieldParameter(field="_id", value=resource_ids, operator=QueryOperators.in_)]
-            )
-
-            query_resources = self.query(query_parameters=params).all().resources
-            resources.extend(query_resources)
-        return resources
-
     def _get_missing_resources(self, resources: List[Union[Resource, FHIRAbstractModel]]):
         missing = check_missing_references(resources)
 
